/*
 * NAME:
 *	netiob - stdio like buffer management
 *
 * SYNOPSIS:
 *	#include <netiob.h>
 *
 *	int	net_setvbuf(NETIOB *p,
 *				char *buf, int	type, size_t  size);
 *	NETIOB *	net_getiob(int	fd, int force, int w);
 *	NETIOB *	net_fgetiobt(FILE *fp, int (*test)(int));
 *	NETIOB *	net_fgetiob(FILE *fp);
 *	int	net_filbuf(NETIOB *p);
 *	int	net_flsbuf(int	c, NETIOB *p);
 *	int	net_fflush(FILE *fp);
 *	int	net_fclose(FILE *fp);
 *
 *
 * DESCRIPTION:
 *	These routines simply allow encryption hooks to be used with
 *	applications that use stdio.
 *
 *	'netio.h' contains macro definitions which override fprintf(),
 *	vfprintf(), fgets(), fgetc(), getc(), putc() and fputc().
 *
 *	Each of the net_*() functions, checks whether the FILE * is
 *	associated with an SSL connection and if not calls the normal
 *	stdio function, otherwise it implements the appropriate
 *	buffered I/O behaviour with net_filbuf() and net_flsbuf()
 *	calling read_net() and write_net() respectively.
 *	
 * SEE ALSO:
 *	rwnet.c
 *
 * AUTHOR:
 *	Simon J. Gerraty <sjg@quick.com.au>
 *	
 *	This is a re-work of a stdio to socket implementation for DOS
 *	I did back in the late 1980s.
 */
/*
 *      @(#)Copyright (c) 1989,1996 Simon J. Gerraty.
 *      
 *      This is free software.  It comes with NO WARRANTY.
 *      Permission to use, modify and distribute this source code 
 *      is granted subject to the following conditions.
 *      1/ that the above copyright notice and this notice 
 *      are preserved in all copies and that due credit be given 
 *      to the author.  
 *      2/ that any changes to this code are clearly commented 
 *      as such so that the author does not get blamed for bugs 
 *      other than his own.
 *      
 *      Please send copies of changes and bug-fixes to:
 *      sjg@quick.com.au
 */
#ifndef lint
static char RCSid[] = "$Id: netiob.c,v 1.7 1997/06/15 04:42:53 sjg Exp $";
#endif

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>

#ifdef USE_SSL
# include "ssl.h"
# include "sslfd.h"
#endif

#include "netiob.h"

#ifndef NUMFDS
# define NUMFDS 64
#endif

#ifndef NEED_NETIOB
FILE *
net_getiob(fd, force, w)
	int fd;
	int force;
	int w;
{
	FILE *fp = stdin;
	int i = 0;
	int m;

	/*
	 * REVISIT:
	 * Unfortunately we cannot rely on any useful indicator
	 * of when we have reached the end of _iob[], or that it
	 * contains anything beyond stderr.  So be very
	 * conservative...
	 */
	m = fd + 1;

	while (i < m) {
		if (fileno(fp) == fd) {
			if (w == (iswrite(fp) > 0))
				return fp;
		}
		++fp;
		++i;
	}
	return 0;
}

FILE *
net_fgetiobt(fp, test)
	FILE *fp;
	int (*test)();
{
	return fp;
}

FILE *
net_fgetiob(fp)
	FILE *fp;
{
	return fp;
}

int
_net_fflush(fp)
	FILE *fp;
{
	return fflush(fp);
}
#else

/*
 * this is the easiest way to handle sockets
 */
static NETIOB *net_iobr[NUMFDS];
static NETIOB *net_iobw[NUMFDS];

int
_net_fflush(p)
	NETIOB *p;
{
	int cnt = -1;
	int n;
	char *cp;
		
	if (!p)
		return 0;

	if ((p->_flag & (_IORW | _IOWRT)) == 0)
		return EOF;
	
	if (p->_file >= 0 && !(p->_flag & (_IOREAD | _IOERR | _IOEOF))) {
		if ((cp = p->_base) == NULL)
			return 0;
		
		n = p->_ptr - p->_base;
		p->_ptr = p->_base;
		p->_cnt = (p->_flag & (_IONBF|_IOLBF)) ? 0 : p->_bufsiz;

		for ( ; n > 0; n -= cnt, cp += cnt) {
			if ((cnt = write_net(p->_file, cp, n)) < 0) {
				p->_flag |= _IOERR;
				return EOF;
			}
		}
	}
	return 0;
}


int
net_setvbuf(p, buf, type, size)
	NETIOB   *p;
	char   *buf;
	int     type;
	size_t  size;
{
	if (!p)
		return -1;
	_net_fflush(p);
	if ((p->_flag & _IOMYBUF) && p->_base != NULL)
		(void) free(p->_base);
	p->_flag &= ~(_IOMYBUF | _IONBF | _IOFBF | _IOLBF);
	switch (type) {
	case _IONBF:
		size = 0;
		buf = NULL;
		break;
	case _IOLBF:
	case _IOFBF:
		if (size == 0)
			size = BUFSIZ;
		if (buf == NULL) {
			buf = (unsigned char *) malloc(size);
			if (buf == NULL)
				return -1;
			p->_flag |= _IOMYBUF;
		}
		break;
	default:
		return -1;
	}
	p->_flag |= type;
	p->_base = p->_ptr = buf;
	p->_cnt = 0;
	p->_bufsiz = size;
	return 0;
}

NETIOB *
net_getiob(fd, force, w)
	int fd;
	int force;
	int w;
{
	static int once = 1;
	NETIOB *p = 0;
	NETIOB **a = (w) ? net_iobw : net_iobr;
	int i;
	
	if (once) {
		memset((char *)net_iobr, 0, sizeof(net_iobr));
		memset((char *)net_iobw, 0, sizeof(net_iobw));
		once = 0;
	}
	if (fd >= 0 && fd < NUMFDS) {
		if ((p = a[fd]) == 0) {
			if (!force)
				return 0;
			
			p = a[fd] = NEW(NETIOB);
			if (p) {
				p->_file = -1;
				p->_base = 0;
				p->_bufsiz = 0;
				p->_flag = 0;
			}
		}
		if (p && p->_file < 0) {
			p->_cnt = 0;
			p->_file = fd;
			p->_flag &= _IOMYBUF;
			p->_ptr = p->_base;
		}
	}
	return p;
}

NETIOB *
net_fgetiobt(fp, test)
	FILE *fp;
	int (*test)();
{
	int fd = (fp) ? fileno(fp) : -1;
	int force = (test && test(fd));

	return net_getiob(fd, force, iswrite(fp));
}

NETIOB *
net_fgetiob(fp)
	FILE *fp;
{
#ifdef USE_SSL
	return net_fgetiobt(fp, (int (*)())ssl_getcon);
#else
	return net_fgetiobt(fp, 0);
#endif
}


int
net_filbuf(p)
	NETIOB *p;
{
	static unsigned char ch;
	register int i = -1;

	if (!p)
		return EOF;
	
	if (p->_file >= 0 && !(p->_flag & (_IOEOF | _IOERR))) {
		if (p->_base == NULL) {
			if (p->_flag & _IONBF) {
				p->_cnt = 0;
				if (read_net(p->_file, &ch, 1) < 0) {
					p->_flag |= (_IOEOF | _IOERR);
					return EOF;
				}
				p->_flag &= ~(_IOEOF | _IOERR);
				return ch;
			} else
				i = net_setvbuf(p, NULL, _IOFBF, 0);
		} else
			i = 0;
	}
	if (i < 0)
		return EOF;

	if ((p->_cnt = read_net(p->_file, p->_base, p->_bufsiz)) <= 0) {
		p->_flag |= (_IOEOF | _IOERR);
		return EOF;
	}
	p->_ptr = p->_base;
	return net_getc(p);
}

int
net_flsbuf(c, p)
	int     c;
	NETIOB 	*p;
{
	static unsigned char ch;
	register int done = 0;
	register int i = -1;
	int n;

	if (!p)
		return EOF;

	p->_flag |= _IOWRT;
	
	if (p->_file >= 0 && !(p->_flag & (_IOERR | _IOEOF))) {
		if (p->_base == NULL) {
			if (p->_flag & _IONBF) {
				p->_cnt = 0;
				ch = (unsigned char) c;
				if (write_net(p->_file, &ch, 1) < 0) {
					p->_flag |= (_IOEOF | _IOERR);
					return EOF;
				}
				return ch;
			} else 
				i = net_setvbuf(p, NULL, _IOFBF, 0);
		} else
			i = 0;
	}
	if (i < 0)
		return EOF;

	n = p->_ptr - p->_base;
	if (n >= p->_bufsiz) {
		if (_net_fflush(p) < 0)
			return EOF;
		n = 0;
	}
	if (n == 0)
		p->_cnt = p->_bufsiz;
	c = net_putc(c, p);
	if (++n == p->_bufsiz || (p->_flag & _IOLBF && c == '\n'))
		if (_net_fflush(p) < 0)
			return EOF;
	return c;
}
#endif /* NEED_NETIOB */

/*
 * PUBLIC functions
 */

int
net_fflush(fp)
	FILE *fp;
{
#ifdef NEED_NETIOB
	NETIOB *p = net_fgetiob(fp);
	
	if (p)
		return _net_fflush(p);
#endif
	return fflush(fp);
}


int
net_flush(fd)
	int fd;
{
	NETIOB *p = net_getiob(fd, 0, 1);

	if (p)
#ifdef NEED_NETIOB
		return _net_fflush(p);
#else
		return fflush(p);
#endif
	return 0;
}


int
net_fclose(fp)
	FILE *fp;
{
#ifdef NEED_NETIOB
	int fd = (fp) ? fileno(fp) : -1;
	NETIOB *p = 0;

	if (p = net_getiob(fd, 0, iswrite(fp))) {
		_net_fflush(p);
#ifdef USE_SSL
		ssl_close(fd);
#endif
		p->_file = -1;
	}
#endif
	return fclose(fp);
}
